/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: channel.c,v 1.28 2002/04/01 11:39:33 lloydy Exp $
 *
 */
#include "server.h"
#include "monkeyircd.h"
#include "channel.h"
#include "utility.h"

/* Linked List Start */
static ChannelInfo list_start;

/* Internal routines */
static int channel_free (ChannelInfo);

/* 
* get a pointer to a specific channel 
*/
ChannelInfo
locate_channel (char *name)
{
  ChannelInfo curr = NULL, prev = NULL;
  int retval = 0;

  if (!list_start)
    {
      return NULL;
    }
  else
    {
      curr = list_start;

      while (curr != NULL && (retval = lc_strncmp (name, curr->name, -1)) < 0)
	{
	  prev = curr;
	  curr = curr->next;
	}

      if (retval != 0)
	return NULL;
    }

  return curr;

}

/* 
* Add a channel to the hash 
*/
int
add_channel (ChannelInfo addme)
{
  ChannelInfo curr = NULL, prev = NULL;

  if (!list_start)
    {
      list_start = addme;
    }
  else
    {
      curr = list_start;
      while (curr != NULL && lc_strncmp (addme->name, curr->name, -1) < 0)
	{
	  prev = curr;
	  curr = curr->next;
	}
      if (curr == list_start)
	{
	  list_start = addme;
	  addme->next = curr;
	}
      else
	{
	  addme->next = prev->next;
	  prev->next = addme;
	}
    }

  channels++;
  return 1;
}

/* 
* Remove a channel from the list
*/
int
remove_channel (ChannelInfo killme)
{
  ChannelInfo curr = NULL, prev = NULL;
  int retval = 0;

  if (!list_start)
    {
      return 0;
    }
  else
    {
      curr = list_start;
      while (curr != NULL
	     && (retval = lc_strncmp (killme->name, curr->name, -1)) < 0)
	{
	  prev = curr;
	  curr = curr->next;
	}

      if (retval == 0)
	{
	  if (curr == list_start)
	    {
	      list_start = curr->next;
	    }
	  else
	    {
	      prev->next = curr->next;
	    }
	  channel_free (curr);
	}
    }
  channels--;
  return 1;
}

/* 
* add a client to a channel client list 
*/
int
channel_add_client (ClientInfo client, ChannelInfo chan, int mode)
{
  ChannelMember curr_user = NULL;

  if (chan->users_start != NULL)
    {
      /* Add at end of list */
      curr_user = chan->users_start;

      /* Find end of channel list */
      while (curr_user->next != NULL)
	{
	  curr_user = curr_user->next;
	}
      /* Add at the position found */
      curr_user->next = ircd_malloc (sizeof (struct _user_wrap));

      curr_user = curr_user->next;
      curr_user->user_info = client;
      curr_user->next = NULL;
      curr_user->chan_mode = mode;
      chan->user_count++;
    }
  else
    {
      /* Add at start of list */
      chan->users_start = ircd_malloc (sizeof (struct _user_wrap));

      curr_user = chan->users_start;
      curr_user->user_info = client;
      curr_user->next = NULL;
      curr_user->chan_mode = mode;
      chan->user_count++;
    }

  return 1;

}


/* 
* remove a client from the channel client list 
*/
int
channel_rem_client (ClientInfo client, ChannelInfo chan)
{
  ChannelMember curr_user = NULL;
  ChannelMember prev_user = NULL;

  curr_user = chan->users_start;

  while (curr_user != NULL)
    {
      if (curr_user->user_info == client)
	{
	  if (curr_user == chan->users_start)
	    {
	      chan->users_start = curr_user->next;
	    }
	  else
	    {
	      prev_user->next = curr_user->next;
	    }
	  free (curr_user);
	  chan->user_count--;
	  return 1;
	}
      else
	{
	  prev_user = curr_user;
	  curr_user = curr_user->next;
	}
    }
  return 0;
}

/* 
* Create a channel 
*/
ChannelInfo
channel_create (char *name, time_t stamp)
{
  ChannelInfo new_chan = NULL;

  /* Get memory for it */
  new_chan = ircd_malloc (sizeof (struct _channel));

  /* Copy Channel name in */
  strncpy (new_chan->name, name, CHANNEL_MAX_LENGTH);
  new_chan->user_limit = -1;

  new_chan->ts = stamp;
  new_chan->topic_ts = time (NULL);

  /* Set initial Mode */
  new_chan->mode = CMODE_TOPIC | CMODE_MEMONLYSEND;

  /* Add channel to hash */
  add_channel (new_chan);

  return new_chan;
}

/* 
* Check the format of a channel name 
*/
int
channel_check (char *chan_name)
{
  int i;
  int chan_len;


  if (chan_name == NULL)
    return 0;

  /* Check first param is correct */
  switch (chan_name[0])
    {

    case '#':
    case '+':
    case '&':
    case '!':
      /* Valid start characters */
      break;

    default:
#ifdef DEBUG
      printf ("Channel %s failed format check\n");
      printf ("Reason: Invalid start character %c\n", chan_name[0]);
#endif
      return 0;
      break;
    }

  chan_len = strlen (chan_name);

  /* Check channel length */
  if (chan_len > CHANNEL_MAX_LENGTH)
    return 0;

  /* Search for invalid characters */
  for (i = 0; i < chan_len; i++)
    {
      switch (chan_name[i])
	{

	case ' ':
	case 7:
	case ',':
	  return 0;
	  break;

	default:
	  break;
	}
    }

  return 1;
}

/* 
* Locate client on a channel 
*/
ChannelMember
channel_is_client_on (ClientInfo client, ChannelInfo chan)
{
  ChannelMember curr_user = NULL;
  int found = 0;

  /* Search for user */
  curr_user = chan->users_start;
  while (curr_user != NULL && found == 0)
    {
      if (curr_user->user_info == client)
	{
	  found = 1;
	}
      else
	{
	  curr_user = curr_user->next;
	}
    }

  /* Return pointer to wrapper or NULL */
  if (found == 1)
    {
      return curr_user;
    }
  else
    {
      return NULL;
    }
}

/*
* Check permissions for sending to a channel
* return 0 if not allowed, and 1 if ok
*/
int
channel_msg_check_perms (ChannelInfo chan, ClientInfo sender)
{
  ChannelMember curr_user = NULL;

  /* Get a pointer if they are on the channel */
  curr_user = channel_is_client_on (sender, chan);

  if (curr_user)
    {
      if ((curr_user->chan_mode & CHANUSR_OP) ||
	  (curr_user->chan_mode & CHANUSR_VOICE) ||
	  (curr_user->chan_mode & CHANUSR_HALFOP))
	{
	  /* They have ops/voice */
	  return 1;
	}
      else
	{
	  if (chan->mode & CMODE_MODERATED)
	    {
	      return 0;
	    }
	}

    }
  else
    {
      if (chan->mode & CMODE_MEMONLYSEND)
	{
	  return 0;
	}
    }
  return 1;
}

/* 
* If we need to close this channel, close it
*/
int
channel_needclose (ChannelInfo chan)
{
  if (chan->users_start == NULL)
    {
      remove_channel (chan);
    }
  return 1;
}

/* 
* Free a channel structure
*/
static int
channel_free (ChannelInfo chan)
{
  int i;

  /* Clear ban list */
  for (i = 0; i < chan->ban_size; i++)
    {
      if (chan->ban_list[i] != NULL)
	free (chan->ban_list[i]);
    }

  free (chan);
  return 1;
}

/*
* send a list of njoins for the channels on this server
*/
int
channels_introduce (ServerInfo s_ptr)
{
  char buffer[BUF_SIZE];
  char start[BUF_SIZE];
  char next[BUF_SIZE];
  char names[BUF_SIZE];
  char *mode = NULL;
  ChannelMember curr_client = NULL;
  ChannelInfo curr = NULL;
  unsigned int avail_chars = 0;

  curr = list_start;
  while (curr)
    {

      /* Print start of SJOIN, with mode and co */
      mode = get_chan_mode_str (curr, '+');
      ircd_sprintf (start, BUF_SIZE, "SJOIN %ld %s %s:", curr->ts,
		    curr->name, mode ? mode : "+ ");

      avail_chars = BUF_SIZE - 1 - strlen (start);
      names[0] = '\0';

      curr_client = curr->users_start;
      while (curr_client)
	{
	  ircd_sprintf (next, BUF_SIZE, "%s%s%s%s ",
			(curr_client->chan_mode & CHANUSR_OP) ? "@" : "",
			(curr_client->chan_mode & CHANUSR_VOICE) ? "+" : "",
			(curr_client->chan_mode & CHANUSR_HALFOP) ? "%" : "",
			curr_client->user_info->nick);
	  if (strlen (next) < avail_chars)
	    {
	      strcat (names, next);
	      avail_chars -= strlen (next);
	    }
	  else
	    {
	      ircd_sprintf (buffer, BUF_SIZE, "%s%s", start, names);
	      send_origin_server (s_ptr, get_my_server ()->name, buffer);
	      strcpy (names, next);
	      avail_chars = BUF_SIZE - 1 - strlen (start) - strlen (next);
	    }
	  curr_client = curr_client->next;
	}
      ircd_sprintf (buffer, BUF_SIZE, "%s%s", start, names);
      send_origin_server (s_ptr, get_my_server ()->name, buffer);

      /* Send the topic */
      ircd_sprintf (buffer, BUF_SIZE, "TOPIC %s %s %ld :%s", curr->name,
		    curr->topic_set[0] == '\0' ? "*" : curr->topic_set,
		    curr->topic_ts, curr->topic);
      send_origin_server (s_ptr, get_my_server ()->name, buffer);

      curr = curr->next;

    }


  return 1;
}

/*
 * Should be able to repeat calls to this
 * first call sets the mask and returns first match
 * any calls with NULL after mask has been set, match using
 * original mask
 */
ChannelInfo
get_channel_with_mask (char *passed_mask)
{
  static char mask[CHANNEL_MAX_LENGTH + 1];
  static ChannelInfo curr = NULL;
  ChannelInfo retval = NULL;
  int match = 0;

  if (passed_mask)
    {
      strncpy (mask, passed_mask, CHANNEL_MAX_LENGTH);
      curr = list_start;
    }

  while (curr && !(match = check_string (curr->name, mask)))
    {
      curr = curr->next;
    }

  if (match)
    {
      retval = curr;
      curr = curr->next;
      return retval;
    }
  else
    {
      return NULL;
    }

  return NULL;
}
