/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: enpress.c,v 1.14 2002/04/01 16:41:04 lloydy Exp $
 *
 */
#include "enpress.h"
#include "utility.h"
#include "logger.h"

/*
 * enpress.c - Deals with encryption and compression of buffers
 */

#ifdef USE_OPENSSL
/*
 * Encrypt buffer using key
 */
int
encrypt_buffer (EncryptInfo info, unsigned char *buffer,
		unsigned char *enc_buffer, int *enc_len)
{
  int enc_tmp_len = 0;
  EVP_CIPHER_CTX encrypt;

  EVP_EncryptInit (&(encrypt), EVP_bf_cbc (), info->key, info->iv);
  EVP_EncryptUpdate (&(encrypt), enc_buffer, enc_len, buffer,
		     strlen ((char *) buffer));
  EVP_EncryptFinal (&(encrypt), enc_buffer + *enc_len, &enc_tmp_len);
  *enc_len += enc_tmp_len;

  EVP_CIPHER_CTX_cleanup(&encrypt);

  return (*enc_len);
}

/*
 * Decrypt buffer using key
 * Assume the buffer is > 512 bytes long
 */
int
decrypt_buffer (EncryptInfo info, unsigned char *buffer, unsigned char *clean,
		int len)
{
  int tmp_len = 0, out_len = 0;
  EVP_CIPHER_CTX decrypt;

  EVP_DecryptInit (&(decrypt), EVP_bf_cbc (), info->key, info->iv);

  if (!EVP_DecryptUpdate (&(decrypt), clean, &out_len, buffer, len))
    {
      printf ("Decryption Failure on call to EVP_DecryptUpdate\n");
    }

  if (!EVP_DecryptFinal (&(decrypt), clean + out_len, &tmp_len))
    {
      printf ("Decryption Failure on call to EVP_DecryptFinal\n");
    }

  out_len += tmp_len;

  EVP_CIPHER_CTX_cleanup(&decrypt);

  return out_len;
}

/*
 * Generate a blowfish key using the text
 */
EncryptInfo
generate_key (unsigned char *key_text)
{
  EncryptInfo new = NULL;

  if (!lc_strncmp ((char *) key_text, "*", -1))
    {
      return NULL;
    }

  new = ircd_malloc (sizeof (struct _encryption_contexts));

  EVP_BytesToKey (EVP_bf_cbc (), EVP_md5 (), NULL, key_text,
		  strlen ((char *) key_text), 1, new->key, new->iv);

  return new;
}

#endif

#ifdef USE_ZLIB
/*
 * Compress a buffer 
 */
int
compress_buffer (unsigned char *buffer, unsigned char *comp_buffer,
		 unsigned int comp_len)
{
  z_stream c_stream;		/* Compression stream */
  int buf_len, err;

  c_stream.zalloc = (alloc_func) NULL;
  c_stream.zfree = (free_func) NULL;
  c_stream.opaque = (voidpf) NULL;

  err = deflateInit (&c_stream, Z_DEFAULT_COMPRESSION);
  CHECK_ERR (err, "deflateInit");

  c_stream.next_in = buffer;
  c_stream.next_out = comp_buffer;

  buf_len = strlen ((char *) buffer) + 1;

  while (c_stream.total_in != (uLong) buf_len
	 && c_stream.total_out < comp_len)
    {
      c_stream.avail_in = c_stream.avail_out = 1;
      err = deflate (&c_stream, Z_NO_FLUSH);
      CHECK_ERR (err, "deflate");
    }

  for (;;)
    {
      c_stream.avail_out = 1;
      err = deflate (&c_stream, Z_FINISH);
      if (err == Z_STREAM_END)
	{
	  break;
	}
      else
	{
	  CHECK_ERR (err, "deflate");
	}
    }

  err = deflateEnd (&c_stream);
  CHECK_ERR (err, "deflateEnd");

  return c_stream.total_out;
}

/*
 * Decompress the buffer
 */
int
uncompress_buffer (unsigned char *compr, unsigned int comprLen,
		   unsigned char *uncompr, unsigned int uncomprLen)
{
  int err;
  z_stream d_stream;		/* decompression stream */

  d_stream.zalloc = (alloc_func) NULL;
  d_stream.zfree = (free_func) NULL;
  d_stream.opaque = (voidpf) NULL;

  d_stream.next_in = compr;
  d_stream.avail_in = 0;
  d_stream.next_out = uncompr;

  err = inflateInit (&d_stream);
  CHECK_ERR (err, "inflateInit");

  while (d_stream.total_out < uncomprLen && d_stream.total_in < comprLen)
    {
      d_stream.avail_in = d_stream.avail_out = 1;	/* force small buffers */
      err = inflate (&d_stream, Z_NO_FLUSH);
      if (err == Z_STREAM_END)
	break;
      CHECK_ERR (err, "inflate");
    }

  err = inflateEnd (&d_stream);
  CHECK_ERR (err, "inflateEnd");

  return d_stream.total_out;
}
#endif
