/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: ircd.c,v 1.71 2002/04/01 16:41:04 lloydy Exp $
 *
 */
#include "monkeyircd.h"
#include "monkeyircd-modules.h"
#include "network.h"
#include "ircd.h"
#include "temp.h"
#include "temp_server.h"
#include "temp_client.h"
#include "resolver.h"

/* Describe The Server State */
int signal_caught = 0;
struct fd_wrap fd_list[MAX_CONNS];
int highest_fd;

/* Utility routines */
static int ircd_init (int[], fd_set *, fd_set *);
static int ircd_sockets_check (int[], fd_set *, fd_set *);
/* Accept Connections */
static int ircd_newconnection (int);
/* Deal with commands */
static int ircd_command (int);
static int ircd_temp_command (struct tempinfo *, Cmd);
static int ircd_regist_command (int, Cmd);
/* Register Clients/Servers */
static int ircd_init_server_connection (struct tempinfo *);
static int ircd_client_register (struct tempinfo *);
static int ircd_server_register (struct tempinfo *);
/* Scan stuff */
static int ircd_check_pings ();
static int ircd_check_resolver ();
static int ircd_check_regist ();
/* Relays numerics if required */
static int ircd_numeric_relay (Cmd);


/* 
* Main entry point 
*/
int
serve_clients (int server_sockets[])
{
  int i;
  struct timeval timeout;
  fd_set connections;
  fd_set serv_connects;
  time_t last_pingscan;
  time_t last_check;

  /* Set everything up */
  ircd_init (server_sockets, &connections, &serv_connects);

  /* The timeout needs setting */
  timeout.tv_sec = 1;
  timeout.tv_usec = 0;

  last_pingscan = time (NULL);
  last_check = time (NULL);

  /* The while loop which is used for serving */
  while (signal_caught == 0)
    {

      /* call select() */
      select (highest_fd + 1, &connections, &serv_connects, NULL, &timeout);

      /* Check sockets for stuff */
      ircd_sockets_check (server_sockets, &connections, &serv_connects);

      /* We only check the resolver and regist stuff every 3 secs */
      if ((last_check + 3) < time (NULL))
	{
	  /* Check Resolver status */
	  ircd_check_resolver ();
	  /* Check registration status */
	  ircd_check_regist ();
	}


      /* Check pings every 30 seconds */
      if ((last_pingscan + 30) < time (NULL))
	{
	  ircd_check_pings ();
	  last_pingscan = time (NULL);
	}

      /* Reset the fd_set */
      FD_ZERO (&connections);
      FD_ZERO (&serv_connects);
      for (i = 0; i < 32 && server_sockets[i] != -1; i++)
	FD_SET (server_sockets[i], &connections);


      for (i = 0; i < MAX_CONNS; i++)
	{
	  if (fd_list[i].temp)
	    {
	      if (fd_list[i].temp->flags & USR_CONNECT)
		{
		  FD_SET (i, &serv_connects);
		}
	    }
	  if (fd_list[i].server)
	    {
	      FD_SET (i, &connections);
	    }

	  if ((fd_list[i].temp) && !(fd_list[i].temp->flags & USR_SHOULDCONN))
	    {
	      FD_SET (i, &connections);
	    }
	  if (fd_list[i].client)
	    {
	      FD_SET (i, &connections);
	    }
	}

      /* Reset highest_fd */
      if (highest_fd == -1)
	{
	  for (i = 0; i < 32 && server_sockets[i] != -1; i++)
	    highest_fd = server_sockets[i];
	}

      /* Reset timeval */
      timeout.tv_sec = 1;
      timeout.tv_usec = 0;
    }

  return 1;
}

/* Check the state of the various sockets in the system */
/* Call appropriate stuff */
static int
ircd_sockets_check (int server_sockets[], fd_set * connections,
		    fd_set * serv_connects)
{
  int i;
  int retval;
  unsigned int rv_size = sizeof (int);

  /* Check for a new connection */
  for (i = 0; i < 32 && server_sockets[i] != -1; i++)
    {
      if (FD_ISSET (server_sockets[i], connections))
	{
	  /* Accept new connection */
	  return ircd_newconnection (server_sockets[i]);
	}
    }

  /* Scan array for client/server/temp pointer */
  for (i = 0; i < MAX_CONNS; i++)
    {
      if (FD_ISSET (i, connections))
	{
	  /* Deal with it */
	  return ircd_command (i);
	}
      else
	{
	  /* We've returned due to an event on a pending CONNECT */
	  if (FD_ISSET (i, serv_connects))
	    {
	      log_msg (LOG_DEBUG2, "Running getsockopt() on socket %d", i);
	      if (!getsockopt
		  (i, SOL_SOCKET, SO_ERROR, (void *) &retval, &rv_size))
		{
		  if (retval != 0)
		    {
		      log_msg (LOG_DEBUG2,
			       "Connect call failed, with error %s",
			       strerror (retval));
		      /* Remove flags */
		      temp_rem (i);
		      fd_list[i].temp = NULL;
		      close (i);
		      return 0;
		    }
		  else
		    {
		      /* The socket is connected, init the connection */
		      log_msg (LOG_DEBUG2,
			       "Connected on fd %d, calling init_server_connection");
		      return ircd_init_server_connection (fd_list[i].temp);
		    }
		}
	      else
		{
		  return 0;
		}
	    }
	}
    }

  return 1;
}

/*
* Init a server connection
*/
static int
ircd_init_server_connection (struct tempinfo *t_ptr)
{
  char buffer[BUF_SIZE];
  struct cline *cline = NULL;
#ifdef USE_OPENSSL
  EncryptInfo ctx = NULL;
#endif

  /* Get the password for this server */
  cline = get_cline (t_ptr->name);

#ifdef USE_OPENSSL
  /* Send our TS4 capab line */
  ctx = generate_key ((unsigned char *) cline->key);

  if (ctx)
    {
      ircd_sprintf (buffer, BUF_SIZE, "CAPAB :ENCRYPT");
      send_temp (t_ptr, buffer);
      t_ptr->ctx = ctx;
    }
#endif


  if (!cline->password)
    {
      /* Try to connect w/o pass */
    }
  else
    {
      /* Send it */
      ircd_sprintf (buffer, BUF_SIZE, "PASS %s %s%s %s %s", cline->password,
		    "0280", VERSION, "CMIRCD|Nothing", "");
      send_temp (t_ptr, buffer);
    }

  ircd_sprintf (buffer, BUF_SIZE, "CAPAB :TS4");
  send_temp (t_ptr, buffer);

  /* Send our info */
  ircd_sprintf (buffer, BUF_SIZE, "SERVER %s 0 0 :%s", get_my_server ()->name,
		get_my_server ()->info);
  send_temp (t_ptr, buffer);

  t_ptr->flags |= USR_SENTMYINFO;
  t_ptr->flags &= ~USR_CONNECT;
  t_ptr->flags |= USR_PASS;

  return 1;
}


/* Accept a new connection, add to temp list */
static int
ircd_newconnection (int server_sock)
{
  struct sockaddr_in6 client_info;
  int client_fd = -1;
  socklen_t sa_size = sizeof (struct sockaddr_in6);
  char buffer[200];
  char ip_buffer[INET6_ADDRSTRLEN];

  /* Accept the connection */
  client_fd =
    accept (server_sock, (struct sockaddr *) &client_info, &sa_size);

  if (client_fd == -1)
    {
      printf ("Failed to accept the connection\n");
      perror ("accept");
      return 0;
    }

  /* Set non-blocking */
  fcntl (client_fd, F_SETFL, O_NONBLOCK);

  /* Update highest fd */
  if (client_fd > highest_fd)
    highest_fd = client_fd;

  /* Call temp_add */
  /* Add to fd array */
  memset (ip_buffer, 0, INET6_ADDRSTRLEN);
  if (client_info.sin6_family == AF_INET6)
    {
      struct sockaddr_in6 *sin6_info = NULL;

      sin6_info = (struct sockaddr_in6 *) &(client_info);

      inet_ntop (AF_INET6, &(sin6_info->sin6_addr), ip_buffer,
		 INET6_ADDRSTRLEN);
    }
  if (client_info.sin6_family == AF_INET)
    {
      struct sockaddr_in *sin_info = NULL;

      sin_info = (struct sockaddr_in *) &(client_info);

      inet_ntop (AF_INET, &(sin_info->sin_addr), ip_buffer, INET6_ADDRSTRLEN);
    }
  printf ("Got IP %s\n", ip_buffer);
  fd_list[client_fd].temp = temp_add (ip_buffer, client_fd);

  if (fd_list[client_fd].temp == NULL)
    {
      /* Shite!!! */
      printf ("Failed to add client into temp array\n");
      printf ("Line %d in ircd.c\n", __LINE__);
      /* Drop client */
      close (client_fd);
      return 0;
    }

  fd_list[client_fd].temp->lastrecv = time (NULL);

  /* XXX - Fixme - IP6 lookups */
  if (client_info.sin6_family == AF_INET)
    {
      struct sockaddr_in *sin_info = NULL;
      sin_info = (struct sockaddr_in *) &client_info;
      fd_list[client_fd].res_id =
	resolver_submit_reverse (sin_info->sin_addr.s_addr);
    }
  if (client_info.sin6_family == AF_INET6)
    {
      fd_list[client_fd].temp->flags |= USR_RESOLVED;
    }


  return 1;

}

/* 
 * Recieve data, parse it, and decide what to do with it 
 */
static int
ircd_command (int client_fd)
{
  char buffer[2048];
  int numbytes, i;
  Cmd cmd = NULL;

  errno = 0;

  memset (buffer, 0, 2048);

  while ((numbytes = recv (client_fd, buffer, 2048, MSG_PEEK)) > 0)
    {

#ifdef USE_OPENSSL
      /* Check to see if we're working on an encrypted socket */
      if ((fd_list[client_fd].server && fd_list[client_fd].server->ctx) ||
	  (fd_list[client_fd].temp && fd_list[client_fd].temp->ctx))
	{
	  char enc_len[5];
	  int len, clean_len;
	  unsigned char enc_buffer[1024];
	  EncryptInfo e_info = NULL;

	  /* Get the right encryption info */
	  e_info =
	    fd_list[client_fd].server ? fd_list[client_fd].server->
	    ctx : fd_list[client_fd].temp->ctx;

	  /* First four bytes contain the number of chars we want */
	  memset (enc_len, 0, 4);
	  if (recv (client_fd, enc_len, 4, 0) != 4)
	    {
	      printf ("Couldn't get the four bytes for the length\n");
	    }


	  enc_len[4] = '\0';
	  log_msg (LOG_DEBUG1, "Read in %s, for length buffer", enc_len);
	  len = atoi (enc_len);
	  log_msg (LOG_DEBUG1, "Need to read %d bytes", len);

	  if (len)
	    {
	      memset (enc_buffer, 0, 1024);
	      recv (client_fd, enc_buffer, len, 0);

	      memset (buffer, 0, 2048);
	      clean_len =
		decrypt_buffer (e_info, enc_buffer, (unsigned char *) buffer,
				len);
	      buffer[clean_len] = '\0';

	      for (i = 0; buffer[i] != '\0'; i++)
		{
		  if (buffer[i] == '\n')
		    {
		      buffer[i] = '\0';
		    }
		}
	    }
	  else
	    {
	      return 0;
	    }

	}
      else
	{
#endif
	  for (i = 0; buffer[i] != '\n' && i < numbytes; i++);

	  /* Actually recieve the data */
	  numbytes = recv (client_fd, buffer, i + 1, 0);
	  buffer[numbytes] = '\0';
#ifdef USE_OPENSSL
	}
#endif

      /* Now buffer contains the decrypted string, or the normal string */

      /* Flood protection */
      if (fd_list[client_fd].client != NULL && numbytes > 0)
	{

	  if (fd_list[client_fd].client->flood_timer < time (NULL))
	    {
	      fd_list[client_fd].client->flood_timer = time (NULL);
	    }
	  else
	    {
	      fd_list[client_fd].client->flood_timer += 2;
	    }

	  if (fd_list[client_fd].client->flood_timer > time (NULL) + 30)
	    {
	      /* They're heavily flooding us, drop them */
	      force_quit (client_fd, "Excess Flood");
	      return 0;
	    }

	  if (fd_list[client_fd].client->flood_timer > time (NULL) + 15)
	    {
	      /* Slight flood, ignore them */
	      return 0;
	    }
	}

      /* Touch the idle timers */
      if (fd_list[client_fd].client)
	{
	  fd_list[client_fd].client->lastrecv = time (NULL);
	}
      if (fd_list[client_fd].server)
	{
	  fd_list[client_fd].server->lastrecv = time (NULL);
	}
      if (fd_list[client_fd].temp)
	{
	  fd_list[client_fd].temp->lastrecv = time (NULL);
	}

      /* Call the Parser */
      log_msg (LOG_DEBUG2, "Received %s", buffer);
      cmd = parse_command (buffer);

      /* Clean up the buffer */
      memset (buffer, 0, 2048);

      if (cmd)
	{
	  if (fd_list[client_fd].temp)
	    {
	      /* Clients in the temp list are restricted */
	      return ircd_temp_command (fd_list[client_fd].temp, cmd);
	    }
	  else
	    {
	      return ircd_regist_command (client_fd, cmd);
	    }
	}

    }

end_of_recv:
  ;



  /* If we had a socket error, kill client */
  if (numbytes <= 0 || (errno))
    {

      /* This is wierd, handle connection reset */
      if (numbytes == 0)
	{
	  errno = ECONNRESET;
	}
      force_quit (client_fd, strerror (errno));
    }

  /* Else feck off */
  return 0;



}

/* 
* Deal with a command from a registered client/server 
*/
static int
ircd_regist_command (int fd, Cmd cmd)
{
  char buffer[BUF_SIZE];

  /* Unload is a special command */
  if (lc_strncmp (cmd->command, "unload", -1) == 0)
    {
      cmd_unload (fd_list[fd].client, fd_list[fd].server, cmd);
      cmd_free (cmd);
      return 1;
    }

  /* So is REHASH */
  if (lc_strncmp (cmd->command, "rehash", -1) == 0)
    {
      cmd_rehash (fd_list[fd].client, fd_list[fd].server, cmd);
      cmd_free (cmd);
      return 1;
    }

  /* Do other commands */
  if (ircd_module_dlopen (fd_list[fd].client, fd_list[fd].server, cmd))
    {
      cmd_free (cmd);
      return 1;
    }

  if (fd_list[fd].server)
    {
      /* Try and relay a numeric */
      return ircd_numeric_relay (cmd);
    }

  /* Invalid command return */
  if (fd_list[fd].client)
    {
      /* Shouldn't get one from another server */
      ircd_sprintf (buffer, BUF_SIZE, "%s :Unknown Command", cmd->command);
      send_numeric (fd_list[fd].client, get_my_server (), 421, buffer);
    }

  cmd_free (cmd);
  return 1;

}


/* 
* Deal with a command from a client in the temp list 
*/
static int
ircd_temp_command (struct tempinfo *temp_ptr, Cmd cmd)
{
  char buffer[BUF_SIZE];
  char *hostname = NULL;

  if (lc_strncmp (cmd->command, "nick", -1) == 0)
    {
      temp_nick (temp_ptr, cmd);
      cmd_free (cmd);
      return 1;
    }
  if (lc_strncmp (cmd->command, "user", -1) == 0)
    {
      temp_user (temp_ptr, cmd);
      cmd_free (cmd);
      return 1;
    }
  if (lc_strncmp (cmd->command, "quit", -1) == 0)
    {
      temp_quit (temp_ptr, cmd);
      cmd_free (cmd);
      return 1;
    }

  /* Server registration commands */
  /* Check resolver, so that the reverse works better */
  if (fd_list[temp_ptr->fd].res_id != -1)
    {
      hostname = resolver_check (fd_list[temp_ptr->fd].res_id);
      if (hostname)
	{
	  strncpy (temp_ptr->host, hostname, HOST_LENGTH);
	  temp_ptr->flags |= USR_RESOLVED;
	}
    }

  if (lc_strncmp (cmd->command, "server", -1) == 0)
    {
      temp_server (temp_ptr, cmd);
      cmd_free (cmd);
      return 1;
    }
  if (lc_strncmp (cmd->command, "pass", -1) == 0)
    {
      temp_pass (temp_ptr, cmd);
      cmd_free (cmd);
      return 1;
    }
  if (lc_strncmp (cmd->command, "capab", -1) == 0)
    {
      temp_capab (temp_ptr, cmd);
      cmd_free (cmd);
      return 1;
    }

  /* Not registered yet error message */
  ircd_sprintf (buffer, BUF_SIZE, "451 %s :You have not registered",
		(temp_ptr->nick[0] == '\0') ? "*" : temp_ptr->nick);
  send_temp (temp_ptr, buffer);

  cmd_free (cmd);
  return 0;
}

/* 
* register a local client 
*/
static int
ircd_client_register (struct tempinfo *temp)
{
  ClientInfo client = NULL;
  char buffer[BUF_SIZE];
  char mode[10];
  int i;

  /* Check we have the flags we need */
  if (!(temp->flags & USR_NICK) || !(temp->flags & USR_USER)
      || !(temp->flags & USR_RESOLVED))
    {
      return 0;
    }

  /* Now, we can actually move them */
  /* Remove from the fd_list */
  fd_list[temp->fd].temp = NULL;

  /* Allocate full clientinfo */
  client = ircd_malloc (sizeof (struct _clientinfo));

  /* Allocate memory for client list */
  client->chanlist_no = 10;
  client->chan_list = calloc (10, sizeof (ChannelInfo));
  if (client->chan_list == NULL)
    {
      printf ("Failed to allocate memory for a channel list\n");
      close (temp->fd);
      return 0;
    }
  /* Zero out chan list */
  for (i = 0; i < client->chanlist_no; i++)
    {
      client->chan_list[i] = NULL;
    }

  /* Setup newly allocated memory */
  client->fd = temp->fd;
  client->mode = temp->mode;
  client->flags = temp->flags;
  client->ts = time (NULL);
  strncpy (client->nick, temp->nick, NICK_LENGTH);
  client->nick[NICK_LENGTH] = '\0';
  strncpy (client->host, temp->host, HOST_LENGTH);
  client->host[HOST_LENGTH] = '\0';
  strncpy (client->user, temp->user, USER_LENGTH);
  client->user[USER_LENGTH] = '\0';
  strncpy (client->real_name, temp->real_name, REAL_NAME_LENGTH);
  client->real_name[REAL_NAME_LENGTH] = '\0';
  client->lastrecv = temp->lastrecv;
  client->idle_timer = time (NULL);
  client->server = get_my_server ();

  /* Add into tree */
  client_add_by_struct (client);
  get_my_server ()->client_count++;
  /* Set pointer in the fd_list */
  fd_list[client->fd].client = client;

  /* Remove temp from the list */
  temp = NULL;
  temp_rem (client->fd);

  /* Send info to client */
  ircd_sprintf (buffer, BUF_SIZE, ":Welcome to the Internet Relay Network %s",
		client->nick);
  send_numeric (client, client->server, 1, buffer);
  ircd_sprintf (buffer, BUF_SIZE, ":Your host is %s, running version %s%s",
		client->server->name, VERSION_TAG, VERSION);
  send_numeric (client, client->server, 2, buffer);
  ircd_sprintf (buffer, BUF_SIZE, ":This server was created on %s",
		load_date);
  send_numeric (client, client->server, 3, buffer);
  ircd_sprintf (buffer, BUF_SIZE, "%s %s%s %s %s", client->server->name,
		VERSION_TAG, VERSION, USERMODES, CHANMODES);
  send_numeric (client, client->server, 4, buffer);

  /* Send LUSERS replies */
  send_lusers (client, "*");

  /* Send MOTD */
  send_motd (client);

  /* Set and report mode */
  if (client->mode != 0)
    {
      /* Find out which modes are set */
      switch (client->mode & (MODE_INVIS | MODE_WALLOPS))
	{
	case (MODE_INVIS | MODE_WALLOPS):
	  ircd_sprintf (mode, 10, "+iw");
	  ircd_sprintf (buffer, 200, "MODE %s :%s", client->nick, mode);
	  send_origin_client (client, client->nick, buffer);
	  break;

	case (MODE_INVIS):
	  ircd_sprintf (mode, 10, "+i");
	  ircd_sprintf (buffer, 200, "MODE %s :+%s", client->nick, mode);
	  send_origin_client (client, client->nick, buffer);
	  break;

	case (MODE_WALLOPS):
	  ircd_sprintf (mode, 10, "+w");
	  ircd_sprintf (buffer, 200, ":%s  MODE %s :+%s", client->nick,
			client->nick, mode);
	  send_origin_client (client, client->nick, buffer);
	  break;

	default:
	  break;
	}
    }

  /* Inform servers */
  ircd_sprintf (buffer, BUF_SIZE, "NICK %s 0 %ld %s %s %s %s 0 :%s",
		client->nick, client->ts, get_mode_str (client->mode),
		client->user, client->host, client->server->name,
		client->real_name);
  servers_announce (NULL, buffer);
  return 1;
}

/* 
* Register a server 
*/
static int
ircd_server_register (struct tempinfo *t_ptr)
{
  char buffer[BUF_SIZE];
  ServerInfo server = NULL;

  /* Check Flags */
  /* Only allow clients which haven't used USER/NICK Commands */
  log_msg (LOG_DEBUG2, "Attempting to register a server on fd %d", t_ptr->fd);
  if (!(t_ptr->flags & USR_SERVER) || !(t_ptr->flags & USR_PASS)
      || (t_ptr->flags & USR_NICK) || (t_ptr->flags & USR_USER))
    {
      return 0;
    }

  fd_list[t_ptr->fd].temp = NULL;

  /* Allocate memory for new server */
  server = ircd_malloc (sizeof (struct _serverinfo));

  /* Copy stuff into new structure */
  server->hops = 1;
  server->fd = t_ptr->fd;
#ifdef USE_OPENSSL
  server->ctx = t_ptr->ctx;
#endif
  strncpy (server->name, t_ptr->name, SERVNAME_MAX);
  server->name[SERVNAME_MAX] = '\0';
  server->lastrecv = t_ptr->lastrecv;
  server->via_ptr = get_my_server ();
  strncpy (server->info, t_ptr->info, SERVINFO_MAX);
  server->info[SERVINFO_MAX] = '\0';
  strncpy (server->id, t_ptr->id, SERVID_MAX);
  server->id[SERVID_MAX] = '\0';

  /* Add to server hash */
  if (!server_hash_add (server))
    {
      log_msg (LOG_BASIC, "Failed to add server to hash");
      free (fd_list[t_ptr->fd].server);
      return 0;
    }

  /* Update the counters */
  get_my_server ()->server_count++;

  /* Send them our information */
  if (!(t_ptr->flags & USR_SENTMYINFO))
    {
      ircd_sprintf (buffer, BUF_SIZE, "SERVER %s 0 0 :%s", server_name,
		    server_info);
      send_origin_server (server, NULL, buffer);
    }

  /* Update fd list */
  fd_list[server->fd].server = server;

  /* Free temporary Structure */
  temp_rem (t_ptr->fd);
  t_ptr = NULL;

  /* Send SVINFO */
  ircd_sprintf (buffer, BUF_SIZE, "SVINFO %d %d %d :%ld", 4, 4, 1,
		time (NULL));
  send_origin_server (server, NULL, buffer);

  /* Log it */
  log_msg (LOG_BASIC, "New Local Server Connection to %s - %s", server->name,
	   server->info);

  /* Introduce the servers we know */
  servers_introduce (server);

  /* Introduce the clients we know */
  clients_introduce (server);

  /* introduce the channels using NJOIN */
  channels_introduce (server);

  /* Relay registration to other servers */
  ircd_sprintf (buffer, BUF_SIZE, "SERVER %s 0 0 :%s", server->name,
		server->info);
  servers_relay (server, get_my_server ()->name, buffer);

  ircd_sprintf (buffer, BUF_SIZE, "NetBurst completed with server %s",
		server->name);
  clients_servnotice (buffer);

  return 1;
}


/* 
* Zeros everything out, and calls various init stuff 
*/
static int
ircd_init (int server_sockets[], fd_set * connections, fd_set * serv_connects)
{
  int i;

  /* Initialise Resolver */
  resolver_init ();

  /* Initialise server hash table */
  server_hash_init (server_name, server_info, server_id);

  assert (server_locate (server_name) != NULL);

  /* Initialise our array of fd's */
  for (i = 0; i < MAX_CONNS; i++)
    {
      fd_list[i].server = NULL;
      fd_list[i].client = NULL;
      fd_list[i].temp = NULL;
      fd_list[i].res_id = -1;
    }

  /* Set highest fd */
  for (i = 0; i < 32 && server_sockets[i] != -1; i++)
    {
      highest_fd = server_sockets[i];
    }

  /* Zero out fd_set */
  FD_ZERO (connections);
  FD_ZERO (serv_connects);
  for (i = 0; i < 32 && server_sockets[i] != -1; i++)
    {
      FD_SET (server_sockets[i], connections);
    }

  return 1;
}

/* 
* Check Pings 
*/
static int
ircd_check_pings ()
{
  int i;
  time_t curr_lastrecv = 0;
  int *flags = NULL;


  for (i = 0; i < MAX_CONNS; i++)
    {
      if (fd_list[i].temp != NULL || fd_list[i].client != NULL
	  || fd_list[i].server != NULL)
	{
	  if (fd_list[i].temp)
	    {
	      curr_lastrecv = fd_list[i].temp->lastrecv;
	      flags = &(fd_list[i].temp->flags);
	    }
	  if (fd_list[i].client)
	    {
	      curr_lastrecv = fd_list[i].client->lastrecv;
	      flags = &(fd_list[i].client->flags);
	    }
	  if (fd_list[i].server)
	    {
	      curr_lastrecv = fd_list[i].server->lastrecv;
	      flags = &(fd_list[i].server->flags);
	    }
	  /* We have a client/server on this fd */
	  if (curr_lastrecv + (2 * PING_SECS) < time (NULL) &&
	      (*flags & USR_PINGSENT))
	    {
	      /* Kill the client, more than 180 secs idle */
	      force_quit (i, "Ping Timeout");
	    }
	  else
	    {
	      /* Check to see if we need to ping them */
	      if (curr_lastrecv + PING_SECS < time (NULL) &&
		  !(*flags & USR_PINGSENT))
		{
		  /* Send a ping */
		  *flags |= USR_PINGSENT;
		  if (fd_list[i].server)
		    {
		      send_ping (&fd_list[i]);
		    }
		}
	    }
	}
    }

  return 1;

}

/* 
* Check resolver status on every socket 
* Copy if complete 
*/
static int
ircd_check_resolver ()
{
  int i;
  char *hostname = NULL;
  struct sockaddr_in sin_addr;

  for (i = 0; i < MAX_CONNS; i++)
    {
      if (fd_list[i].res_id != -1)
	{
	  hostname = resolver_check (fd_list[i].res_id);
	  if (hostname)
	    {
	      /* We only need to use temp to resolve */
	      /* Anything else should be resolved */
	      if (fd_list[i].temp)
		{
		  if (fd_list[i].temp->flags & USR_SHOULDCONN)
		    {
		      log_msg (LOG_DEBUG2, "Running Connect() on fd %d", i);
		      log_msg (LOG_DEBUG2, "IP is %s", hostname);
		      /* Its a pending server connect */
		      /* Connect on it */
		      sin_addr.sin_family = AF_INET;
		      /* Copy IP to numeric dest, and into server structure */
		      inet_aton (hostname, &(sin_addr.sin_addr));
		      fd_list[i].temp->connect_ip.s_addr =
			sin_addr.sin_addr.s_addr;
		      sin_addr.sin_port = htons (fd_list[i].temp->port);
		      memset (&(sin_addr.sin_zero), 0, 8);
		      connect (i, (struct sockaddr *) &sin_addr,
			       sizeof (struct sockaddr));
		      fd_list[i].temp->flags &= ~USR_SHOULDCONN;
		      fd_list[i].temp->flags |= USR_CONNECT;
		    }
		  else
		    {
		      strncpy (fd_list[i].temp->host, hostname, HOST_LENGTH);
		    }
		  fd_list[i].temp->flags |= USR_RESOLVED;

		}
	      /* Not sure we need it */
	      if (fd_list[i].server)
		{
		  strncpy (fd_list[i].server->host, hostname, HOST_LENGTH);
		}
	      fd_list[i].res_id = -1;
	      hostname = NULL;
	    }
	}

    }

  return 1;
}

/*
* Check to see if we need to register clients 
*/
static int
ircd_check_regist ()
{
  int i;

  for (i = 0; i < MAX_CONNS; i++)
    {
      if (fd_list[i].temp != NULL)
	{
	  if (ircd_client_register (fd_list[i].temp))
	    {

	    }
	  else
	    {
	      ircd_server_register (fd_list[i].temp);
	    }
	}
    }

  return 1;
}


/*
* Relay numerics to the destination client
* Only relay 100+ in codes
*/
static int
ircd_numeric_relay (Cmd cmd)
{
  int code, i;
  ClientInfo d_ptr = NULL;
  ServerInfo s_ptr = NULL;
  /* Use the real max size now */
  char buffer[512];

  /* Just return if its got no origin */
  if (!cmd->origin)
    {
      cmd_free (cmd);
      return 0;
    }

  /* Check it is from a server */
  s_ptr = server_locate (cmd->origin);
  if (!s_ptr)
    {
      cmd_free (cmd);
      return 0;
    }

  /* Check the numeric is between 100 and 999 inclusive */
  code = atoi (cmd->command);
  if (code < 100 || code > 999)
    {
      cmd_free (cmd);
      return 0;
    }

  /* Now get the dest */
  d_ptr = client_locate (cmd->params[0]);

  if (!d_ptr)
    {
      cmd_free (cmd);
      return 0;
    }

  ircd_sprintf (buffer, 490, "%i %s", code, cmd->params[0]);
  for (i = 1; cmd->params[i]; i++)
    {
      ircd_sprintf (buffer, 490, "%s %s", buffer, cmd->params[i]);
    }
  ircd_sprintf (buffer, 490, "%s :%s", buffer, cmd->text);

  send_origin_client (d_ptr, s_ptr->name, buffer);

  cmd_free (cmd);
  return 1;
}
