/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: temp_client.c,v 1.15 2002/04/01 11:39:33 lloydy Exp $
 *
 */
#include "temp_client.h"
#include "ircd.h"

/* temp_client.c - Commands valid in temporary unregistered state,
* related to client operation
* Server-related commands (including PASS) are in temp_server.c
*/

/* 
* Inform client if nick is in use 
* will never be able to use nick in the struct
*/
static int
tmp_nick_taken (struct tempinfo *t_ptr, char *requested_nick)
{
  char buffer[200];

  ircd_sprintf (buffer, 200, "433 * %s :Nickname already in use",
		requested_nick);
  return send_temp (t_ptr, buffer);
}

/* 
* Handles NICK in temp list 
*/
int
temp_nick (struct tempinfo *t_ptr, Cmd cmd)
{
  ClientInfo exist = NULL;
  char *new_nick = NULL;
  char buffer[100];
  char actual_nick[NICK_LENGTH + 1];

  /* Send AUTH message */
  if (!(t_ptr->flags & USR_SENTINITAUTH))
    {
      send_temp (t_ptr, "NOTICE * :*** Looking Up Hostname");
      t_ptr->flags |= USR_SENTINITAUTH;
    }

  /* cHeck we actually have a nick */
  if (cmd->params[0] || cmd->text)
    {
      /* Let it use either text or params[0] */
      if (!cmd->params[0])
	{
	  new_nick = cmd->text;
	}
      else
	{
	  new_nick = cmd->params[0];
	}
      /* Check validity of new nick */
      if (!check_nick (new_nick))
	{
	  ircd_sprintf (buffer, 100, "432 * %s :Erroneous Nick Name",
			new_nick);
	  send_temp (t_ptr, buffer);
	  return 0;
	}
      strncpy (actual_nick, new_nick, NICK_LENGTH);
      exist = client_locate (actual_nick);
      if (!exist)
	{
	  /* User isn't in tree */
	  if (!temp_locate (actual_nick))
	    {
	      strncpy (t_ptr->nick, actual_nick, NICK_LENGTH);
	      t_ptr->nick[NICK_LENGTH] = '\0';
	      t_ptr->flags |= USR_NICK;
	      return 1;
	    }
	  else
	    {
	      /* User is in list */
	      /* Inform User */
	      tmp_nick_taken (t_ptr, actual_nick);
	      return 0;
	    }
	}
      else
	{
	  /* In Tree */
	  tmp_nick_taken (t_ptr, actual_nick);
	  return 0;
	}
    }
  else
    {
      return 0;
    }


}


/* 
* deal with quit in the temp list 
*/
int
temp_quit (struct tempinfo *t_ptr, Cmd cmd)
{
  int i;

  if (!cmd)
    return 0;

  /* Remove it from the lsit */
  fd_list[t_ptr->fd].temp = NULL;

  /* Cancel any dns request */
  resolver_cancel (fd_list[t_ptr->fd].res_id);
  fd_list[t_ptr->fd].res_id = -1;

  if (t_ptr->fd == highest_fd)
    {
      for (i = highest_fd; i >= 0; i--)
	{
	  if (fd_list[i].server != NULL || fd_list[i].client != NULL
	      || fd_list[i].temp != NULL)
	    {
	      highest_fd = i;
	      i = -1;
	    }
	}
      if (i == 0)
	highest_fd = -1;
    }

  /* Close the socket */
  close (t_ptr->fd);

  /* Remove the thing from the list */
  temp_rem (t_ptr->fd);

  return 1;
}

/*
* deal with USER in temp list
*/
int
temp_user (struct tempinfo *sender, Cmd cmd)
{
  int supp_mode;
  char buffer[BUF_SIZE];

  if (!check_params (cmd, 2))
    {
      if (sender->nick[0] == '\0')
	{
	  ircd_sprintf (buffer, BUF_SIZE, "461 * USER :Not enough params");
	  send_temp (sender, buffer);
	}
      else
	{
	  ircd_sprintf (buffer, BUF_SIZE, "461 %s USER :Not Enough Params",
			sender->nick);
	  send_temp (sender, buffer);
	}
      return 1;
    }

  if (cmd->params[0] != NULL)
    {
      if (!check_nick (cmd->params[0]))
	{
	  strncpy (sender->user, sender->nick, NICK_LENGTH);
	  sender->user[NICK_LENGTH] = '\0';
	}
      else
	{
	  strncpy (sender->user, cmd->params[0], NICK_LENGTH);
	  sender->user[NICK_LENGTH] = '\0';
	}
    }

  if (cmd->text != NULL)
    {
      strncpy (sender->real_name, cmd->text, REAL_NAME_LENGTH);
      sender->real_name[REAL_NAME_LENGTH] = '\0';
    }

  /* convert mode in cmd->params[1] to an int */
  /* mask it and add modes to cmd->mode   */
  /* _don't_ just convert and add to struct */
  sender->mode = 0;
  if (cmd->params[1] != NULL)
    {
      supp_mode = atoi (cmd->params[1]);
    }
  else
    {
      supp_mode = 0;
    }
  switch (supp_mode)
    {
    case (MODE_INVIS | MODE_WALLOPS):
      sender->mode = MODE_INVIS | MODE_WALLOPS;
      break;
    case MODE_INVIS:
      sender->mode = MODE_INVIS;
      break;
    case MODE_WALLOPS:
      sender->mode = MODE_WALLOPS;
      break;
    default:
      sender->mode = 0;
      break;
    }
  sender->flags |= USR_USER;
  return 1;
}
